# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from swaggerit.exceptions import SwaggerItModelError
from swaggerit.models.orm.factory import FactoryOrmModels
from swaggerit.utils import get_model
from importlib import import_module


class ModuleObjectLoader(object):
    _objects = dict()

    @classmethod
    def load(cls, config):
        key = '{}.{}'.format(config['path'], config['object_name'])
        obj = cls._objects.get(key)

        if obj is None:
            try:
                module = import_module(config['path'])
                obj = getattr(module, config['object_name'])
                cls._objects[key] = obj

            except Exception as error:
                raise SwaggerItModelError(
                    "Error loading module '{}.{}'.\nError Class: {}. Error Message: {}".format(
                        config['path'], config['object_name'],
                        error.__class__.__name__, str(error)
                    )
                )

        return obj


def build_item_key(name, sufix=None):
    name = name.lower().replace(' ', '_')
    if sufix:
        return '{}_{}'.format(name, sufix)
    return name


def get_items_model(engine):
    items_types_model_key = build_item_key(engine['item_type']['name'], engine['store_id'])
    return get_model(items_types_model_key)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.authorizer import MyrecoAuthorizer
from myreco.factory import ModelsFactory
from swaggerit.constants import SWAGGER_VALIDATOR
from swaggerit.aiohttp_api import AioHttpAPI
from swaggerit.exceptions import SwaggerItAPIError
from copy import deepcopy
import ujson


class MyrecoAPI(AioHttpAPI):
    def __init__(self, *, type_='recommender', sqlalchemy_bind=None, redis_bind=None,
                 swagger_json_template=None, title=None, version='1.0.0',
                 get_swagger_req_auth=True, loop=None, debug=False,
                 factory_class=ModelsFactory):
        self.models_factory = factory_class()
        self.all_models = self.models_factory.make_all_models(type_)
        authorizer = MyrecoAuthorizer(self.all_models['users'])
        models = [model for model in self.all_models.values() if hasattr(model, '__api__')]

        AioHttpAPI.__init__(
            self, models,
            sqlalchemy_bind=sqlalchemy_bind,
            redis_bind=redis_bind,
            swagger_json_template=swagger_json_template,
            title=title, version=version,
            authorizer=authorizer,
            get_swagger_req_auth=get_swagger_req_auth,
            loop=loop, debug=debug
        )
        self._set_items_models_routes(self.all_models['items_types'])

    def _set_items_models_routes(self, items_types_model):
        base_uri = '/{items_model_name}'
        handler = self._set_handler_decorator(items_types_model.items_models_handler)
        self._set_route(base_uri, 'GET', handler)
        self._set_route(base_uri, 'POST', handler)
        self._set_route(base_uri, 'PATCH', handler)
        self._set_route(base_uri + '/{item_key}', 'GET', handler)
        self._set_route(base_uri + '/{item_key}', 'POST', handler)

    def update_swagger_paths(self, model):
        SWAGGER_VALIDATOR.validate(model.__schema__)
        new_paths, definitions = self._get_model_paths_and_definitions(model)
        self.swagger_json['paths'].update(new_paths)
        self.swagger_json['definitions'].update(definitions)

    def remove_swagger_paths(self, model):
        new_swagger_json = deepcopy(self.swagger_json)
        paths_to_remove = set()
        definitions_to_remove = set()
        name = model.__name__

        for path in model.__schema__:
            if path != 'definitions':
                new_swagger_json['paths'].pop(path)
                paths_to_remove.add(path)

        for definition in list(new_swagger_json['definitions'].keys()):
            if name in definition:
                new_swagger_json['definitions'].pop(definition)
                definitions_to_remove.add(definition)

        [self.swagger_json['paths'].pop(path) for path in paths_to_remove]
        [self.swagger_json['definitions'].pop(path) for path in definitions_to_remove]



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from swaggerit.utils import get_model_schema
from sqlalchemy.ext.declarative import AbstractConcreteBase
import sqlalchemy as sa
import ujson


class EnginesCoresModelBase(AbstractConcreteBase):
    __tablename__ = 'engines_cores'
    __schema__ = get_model_schema(__file__)

    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.String(255), unique=True, nullable=False)
    configuration_json = sa.Column(sa.Text, nullable=False)

    @property
    def configuration(self):
        if not hasattr(self, '_configuration'):
            self._configuration = ujson.loads(self.configuration_json)
        return self._configuration

    async def _setattr(self, attr_name, value, session, input_):
        if attr_name == 'configuration':
            value = ujson.dumps(value)
            attr_name = 'configuration_json'

        await super()._setattr(attr_name, value, session, input_)

    def _format_output_json(self, dict_inst, schema):
        if schema.get('configuration') is not False:
            dict_inst.pop('configuration_json')
            dict_inst['configuration'] = self.configuration



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from swaggerit.utils import get_model_schema
from swaggerit.json_builder import JsonBuilder
from swaggerit.exceptions import SwaggerItModelError
from myreco.engines.cores.filters.factory import FiltersFactory
from myreco.utils import ModuleObjectLoader, get_items_model
from sqlalchemy.ext.declarative import AbstractConcreteBase, declared_attr
import random as random_
import sqlalchemy as sa
import hashlib
import ujson


class PlacementsModelBase(AbstractConcreteBase):
    __tablename__ = 'placements'
    __schema__ = get_model_schema(__file__)

    hash = sa.Column(sa.String(255), unique=True, nullable=False)
    small_hash = sa.Column(sa.String(255), primary_key=True)
    name = sa.Column(sa.String(255), nullable=False)
    ab_testing = sa.Column(sa.Boolean, default=False)
    show_details = sa.Column(sa.Boolean, default=True)
    distribute_recos = sa.Column(sa.Boolean, default=False)

    @declared_attr
    def store_id(cls):
        return sa.Column(sa.ForeignKey('stores.id'), nullable=False)

    @declared_attr
    def variations(cls):
        return sa.orm.relationship('VariationsModel', uselist=True, passive_deletes=True)

    async def init(self, session, input_=None, **kwargs):
        await super().init(session, input_=input_, **kwargs)
        self._set_hash()

    def _set_hash(self):
        if self.name and not self.hash:
            hash_ = hashlib.new('ripemd160')
            hash_.update(self.name.encode() + bytes(self.store_id))
            self.hash = hash_.hexdigest()

    def __setattr__(self, name, value):
        if name == 'hash':
            self.small_hash = value[:5]

        super().__setattr__(name, value)

    @classmethod
    async def get_recommendations(cls, req, session):
        placement = await cls._get_placement(req, session)
        if placement is None:
            return cls._build_recos_response(None)

        explict_fallbacks = req.query.pop('explict_fallbacks', False)
        input_variables = req.query
        show_details = req.query.pop('show_details', placement.get('show_details'))
        distribute_recos = placement.get('distribute_recos')
        recos = slots = []
        recos_key = 'slots'

        for slot in placement['variations'][0]['slots']:
            slot_recos = {'fallbacks': []}
            slot_recos['main'] = \
                await cls._get_recos_by_slot(slot, input_variables, session, show_details)

            await cls._get_fallbacks_recos(slot_recos, slot, input_variables, session, show_details)

            slot = {'name': slot['name'], 'item_type': slot['engine']['item_type']['name']}
            slot['recommendations'] = slot_recos
            if slot_recos['main'] or slot_recos['fallbacks']:
                slots.append(slot)

        if not slots:
            return cls._build_recos_response(None)

        if not explict_fallbacks:
            for slot in slots:
                slot['recommendations'] = cls._get_all_slot_recos(slot['recommendations'])

            if distribute_recos:
                recos_key = 'distributed_recommendations'
                recos = cls._get_all_recos_from_slots(slots)
                recos = cls._distribute_recos(recos)

        placement = {'name': placement['name'], 'small_hash': placement['small_hash']}
        placement[recos_key] = recos

        return cls._build_recos_response(placement)

    @classmethod
    async def _get_placement(cls, req, session):
        small_hash = req.path_params['small_hash']
        placements = await cls.get(session, {'small_hash': small_hash})

        if not placements:
            return None

        return placements[0]

    @classmethod
    async def _get_recos_by_slot(cls, slot, input_variables, session, show_details, max_recos=None):
        try:
            engine = slot['engine']
            items_model = get_items_model(engine)
            engine_vars, filters = \
                cls._get_variables_and_filters(slot, items_model, input_variables)
            core_config = engine['core']['configuration']['core_module']
            core_instance = ModuleObjectLoader.load(core_config)(engine, items_model)
            max_recos = slot['max_recos'] if max_recos is None else max_recos

            return await core_instance.get_recommendations(
                session, filters, max_recos, show_details, **engine_vars)
        except Exception as error:
            cls._logger.debug('Slot:\n' + ujson.dumps(slot, indent=4))
            cls._logger.debug('Input Variables:\n' + ujson.dumps(input_variables, indent=4))
            raise error

    @classmethod
    def _get_variables_and_filters(cls, slot, items_model, input_variables):
        engine_vars = dict()
        filters = dict()
        engine = slot['engine']

        for engine_var in slot['slot_variables']:
            var_name = engine_var['variable']['name']
            var_engine_name = engine_var['inside_engine_name']

            if var_name in input_variables:
                var_value = input_variables[var_name]

                if engine_var['is_filter']:
                    filter_schema, input_schema = \
                        cls._get_filter_and_input_schema(engine, engine_var)

                    if filter_schema is not None and input_schema is not None:
                        filter_ = FiltersFactory.make(items_model, engine_var, filter_schema)
                        filters[filter_] = JsonBuilder.build(var_value, input_schema)

                else:
                    schema = cls._get_variable_schema(engine, engine_var)
                    if schema is not None:
                        engine_vars[var_engine_name] = JsonBuilder.build(var_value, schema)

        return engine_vars, filters

    @classmethod
    def _get_filter_and_input_schema(cls, engine, engine_var):
        for var in engine['item_type']['available_filters']:
            if var['name'] == engine_var['inside_engine_name']:
                if engine_var['filter_type'].endswith('Of'):
                    input_schema = {'type': 'array', 'items': {'type': 'string'}}

                elif var['schema'].get('type') != 'array':
                    input_schema = {'type': 'array', 'items': var['schema']}

                else:
                    input_schema = var['schema']

                filter_schema = var['schema']
                return filter_schema, input_schema

        return None, None

    @classmethod
    def _get_variable_schema(cls, engine, engine_var):
        for var in engine['variables']:
            if var['name'] == engine_var['inside_engine_name']:
                return var['schema']

    @classmethod
    async def _get_fallbacks_recos(cls, slot_recos, slot, input_variables, session, show_details):
        if len(slot_recos['main']) != slot['max_recos']:
            for fallback in slot['fallbacks']:
                fallbacks_recos_size = \
                    sum([len(fallback) for fallback in slot_recos['fallbacks']])
                max_recos = slot['max_recos'] - len(slot_recos['main']) - fallbacks_recos_size
                if max_recos == 0:
                    break

                fallback_recos = await cls._get_recos_by_slot(
                    fallback, input_variables, session, show_details, max_recos)
                all_recos = cls._get_all_slot_recos(slot_recos)
                fallback_recos = cls._unique_recos(fallback_recos, all_recos)
                slot_recos['fallbacks'].append(fallback_recos)

    @classmethod
    def _get_all_slot_recos(cls, slot_recos):
        all_recos = list(slot_recos['main'])
        [all_recos.extend(fallback_recos) for fallback_recos in slot_recos['fallbacks']]
        return all_recos

    @classmethod
    def _get_all_recos_from_slots(cls, slots):
        recos = []
        for slot in slots:
            for reco in slot['recommendations']:
                reco['type'] = slot['item_type']

            recos.append(slot['recommendations'])
        return recos

    @classmethod
    def _distribute_recos(cls, recos_list, random=True):
        total_length = sum([len(recos) for recos in recos_list])
        total_items = []
        initial_pos = 0

        for recos in recos_list:
            if not recos:
                continue

            step = int(total_length / len(recos))

            if random:
                initial_pos = random_.randint(0, step-1)

            positions = range(initial_pos, total_length, step)
            zip_items_positions = zip(recos, positions)
            total_items.extend(zip_items_positions)

        random_.shuffle(total_items)
        sorted_items = sorted(total_items, key=(lambda each: each[1]))

        return [i[0] for i in sorted_items]

    @classmethod
    def _unique_recos(cls, recos, all_recos):
        unique = list()
        [unique.append(reco) for reco in recos if not all_recos.count(reco)]
        return unique

    @classmethod
    def _build_recos_response(cls, recos):
        if recos is None:
            return cls._build_response(404)
        else:
            headers = {'Content-Type': 'application/json'}
            return cls._build_response(200, body=cls._pack_obj(recos), headers=headers)


class VariationsModelBase(AbstractConcreteBase):
    __tablename__ = 'variations'
    __use_redis__ = False

    id = sa.Column(sa.Integer, primary_key=True)
    weight = sa.Column(sa.Float)

    @declared_attr
    def placement_hash(cls):
        return sa.Column(sa.ForeignKey('placements.hash', ondelete='CASCADE', onupdate='CASCADE'), nullable=False)

    @declared_attr
    def slots(cls):
        return sa.orm.relationship('SlotsModel',
                uselist=True, secondary='variations_slots')


class ABTestUsersModelBase(AbstractConcreteBase):
    __tablename__ = 'ab_test_users'
    __use_redis__ = False

    id = sa.Column(sa.Integer, primary_key=True)

    @declared_attr
    def variation_id(cls):
        return sa.Column(sa.ForeignKey('variations.id'), nullable=False)


def build_variations_slots_table(metadata, **kwargs):
    return sa.Table(
        'variations_slots', metadata,
        sa.Column('variation_id', sa.Integer, sa.ForeignKey('variations.id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        sa.Column('slot_id', sa.Integer, sa.ForeignKey('slots.id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        **kwargs)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.engines.cores.base import EngineCoreBase
from abc import abstractmethod
from glob import glob
import gzip
import os.path
import asyncio


class EngineCoreObjectsExporter(EngineCoreBase):

    @abstractmethod
    def export_objects(self, session):
        pass

    def _build_csv_readers(self, pattern=''):
        readers = []
        pattern = os.path.join(self._data_path, '{}*.gz'.format(pattern))

        for filename in glob(pattern):
            file_ = gzip.open(filename, 'rt')
            readers.append(file_)

        return readers

    async def _get_items_indices_map_dict(self, session):
        items_indices_map_dict = await self._items_indices_map.get_all(session)

        if not items_indices_map_dict.values():
            raise EngineError(
                "The Indices Map for '{}' is empty. Please update these items"
                .format(self.engine['item_type']['name']))

        return items_indices_map_dict

    def _run_coro(self, coro, loop):
        if loop.is_running():
            return asyncio.run_coroutine_threadsafe(coro, loop).result()
        else:
            return loop.run_until_complete(coro)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.engines.cores.top_seller.redis_object import TopSellerRedisObject
from myreco.engines.cores.objects_exporter import EngineCoreObjectsExporter
from myreco.engines.cores.recommender import EngineCoreRecommender


class TopSellerEngineCore(EngineCoreObjectsExporter, EngineCoreRecommender):
    __configuration_schema__ = {
        "type": "object",
        "required": ["days_interval"],
        "properties": {
            "days_interval": {"type": "integer"}
        }
    }

    async def _build_rec_vector(self, session, **variables):
        return await TopSellerRedisObject(self).get_numpy_array(session)

    def export_objects(self, session):
        self._logger.info("Started export objects")

        readers = self._build_csv_readers()
        items_indices_map_dict = self._run_coro(
            self._items_indices_map.get_all(session),
            session.loop
        )

        top_seller = TopSellerRedisObject(self)
        ret = self._run_coro(
            top_seller.update(readers, session, items_indices_map_dict),
            session.loop
        )

        self._logger.info("Finished export objects")
        return ret



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.items_types.items.models_collection import ItemsModelsCollection
from myreco.items_types.base_model import ItemsTypesModelBase
from myreco.utils import ModuleObjectLoader
from swaggerit.exceptions import SwaggerItModelError
from jsonschema import Draft4Validator
from tempfile import NamedTemporaryFile
from aiofiles import zip_open
from io import BytesIO
import ujson
import boto3
import os
import asyncio
import gc


class ItemsModelsCollectionDataImporter(ItemsModelsCollection):

    async def post_import_data_file_job(self, req, session):
        content_type = req.headers.get('content-type')
        if content_type != 'application/zip':
            raise SwaggerItModelError("Invalid content type '{}'".format(content_type))

        stream = BytesIO()
        async for line in req.body:
            stream.write(line)
        stream.seek(0)

        return self._create_job(
            self._run_import_data_file_job,
            req, session, '_importer', stream=stream)

    def _create_job(self, func, req, session, jobs_id_prefix, stream=None):
        session = self._copy_session(session)
        store_id = req.query['store_id']
        items_model = self._get_model(req.query)
        jobs_id = items_model.__key__ + jobs_id_prefix
        return super()._create_job(
            func, jobs_id,
            req, session,
            items_model, store_id,
            stream=stream
        )

    def _run_import_data_file_job(self, req, session, items_model, store_id, stream):
        upload_file = req.query.get('upload_file', True)
        if upload_file:
            self._put_file_on_s3(stream, items_model, session, store_id)
            stream.seek(0)

        result = asyncio.run_coroutine_threadsafe(
            self._update_items_from_zipped_file(stream, items_model, session),
            session.loop
        ).result()

        gc.collect()
        return result

    def _put_file_on_s3(self, stream, items_model, session, store_id):
        self._logger.info("Started put file on S3 for '{}'".format(items_model.__key__))

        store = asyncio.run_coroutine_threadsafe(
            self.get_model('stores').get(session, [{'id': store_id}]),
            session.loop
        ).result()[0]

        s3_bucket = store['configuration']['aws']['s3']['bucket']
        access_key_id = store['configuration']['aws'].get('access_key_id')
        secret_access_key = store['configuration']['aws'].get('secret_access_key')
        s3_key = '{}.zip'.format(items_model.__key__)

        boto3.resource(
            's3',
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key
        ).Bucket(s3_bucket).put_object(Body=stream, Key=s3_key)

        self._logger.info("Finished put file on S3 for '{}'".format(items_model.__key__))

    async def _update_items_from_zipped_file(self, stream, items_model, session):
        tempfile = NamedTemporaryFile(delete=False)
        filename = tempfile.name
        tempfile.write(stream.read())
        tempfile.close()

        feed = await zip_open(filename)
        result = await self._update_items_from_file(feed, items_model, session)

        await feed.close()
        os.remove(filename)
        return result

    async def _update_items_from_file(self, feed, items_model, session):
        self._logger.info("Started update items from file for '{}'".format(items_model.__key__))

        warning_message = "Invalid line for model '{}': ".format(items_model.__key__) + '{}'
        validator = Draft4Validator(items_model.__item_type__['schema'])
        lines = []

        old_keys = set(await session.redis_bind.hkeys(items_model.__key__))
        new_keys = set()
        success_lines = 0
        errors_lines = 0
        empty_lines = 0

        async for line in feed:
            try:
                line = line.strip().decode()
                if not line:
                    empty_lines += 1
                    continue

                line = ujson.loads(line)
                validator.validate(line)

            except:
                errors_lines += 1
                self._logger.warning(warning_message.format(line))
                continue

            else:
                success_lines += 1
                new_keys.add(items_model.get_instance_key(line))
                lines.append(line)

            if len(lines) == 1000:
                await items_model.insert(session, lines)
                lines = []

        if lines:
            self._post_processing_import(lines)
            await items_model.insert(session, lines)

        del lines
        old_keys.difference_update(new_keys)

        if old_keys:
            await session.redis_bind.hdel(items_model.__key__, *old_keys)

        await self._set_stock_filter(session, items_model)

        self._logger.info("Finished update items from file for '{}'".format(items_model.__key__))

        return {
            'success_lines': success_lines,
            'errors_lines': errors_lines,
            'empty_lines': empty_lines
        }

    def _post_processing_import(cls, items):
        post_processing_import = cls.__item_type__['post_processing_import']
        if post_processing_import is not None:
            post_processing_import = ModuleObjectLoader.load(post_processing_import)
            post_processing_import.execute(items)

    async def get_import_data_file_job(self, req, session):
        jobs_id = self._get_model(req.query).__key__ + '_importer'
        return await self._get_job(jobs_id, req, session)



# MIT License

# Copyright (c) 2016 Diogo Dutra <dutradda@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.engines.cores.filters.filters import (BooleanFilterBy, SimpleFilterBy, ObjectFilterBy,
    ArrayFilterBy, SimpleFilterOf, ObjectFilterOf,
    ArrayFilterOf, IndexFilterOf, IndexFilterByPropertyOf)


class FiltersFactory(object):

    @classmethod
    def make(cls, items_model, engine_variable, schema, skip_values=None):
        filter_type = schema['type']
        var_name = engine_variable['inside_engine_name']
        type_ = engine_variable['filter_type']
        is_inclusive = engine_variable['is_inclusive_filter']
        id_names = None

        if type_ == 'Index Of':
            filter_class = IndexFilterOf

        elif type_ == 'Index By Property Of':
            filter_class = IndexFilterByPropertyOf

        elif filter_type == 'integer' or filter_type == 'string':
            if type_ == 'By Property':
                filter_class = SimpleFilterBy
            else:
                filter_class = SimpleFilterOf

        elif filter_type == 'boolean':
            filter_class = BooleanFilterBy

        elif filter_type == 'object':
            id_names = schema['id_names']
            if type_ == 'By Property':
                filter_class = ObjectFilterBy
            else:
                filter_class = ObjectFilterOf

        elif filter_type == 'array':
            if type_ == 'By Property':
                filter_class = ArrayFilterBy
            else:
                filter_class = ArrayFilterOf

        return filter_class(items_model, var_name, is_inclusive, id_names, skip_values)
